
import torch
import torch.nn as nn
from SelfONN import SelfONN1d as SelfONN1dlayer



torch.manual_seed(0)

class ResBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, subsample_length):
        super().__init__()


        if in_channels != out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv1d(in_channels, out_channels, kernel_size = 1, stride=1),
                nn.BatchNorm1d(out_channels),
                nn.ReLU(),
                nn.MaxPool1d(subsample_length)
            )
        else:
            self.shortcut = nn.MaxPool1d(subsample_length)

        self.conv1 = nn.Sequential(
            nn.Conv1d(in_channels, out_channels, kernel_size, stride=subsample_length, padding=(kernel_size//2)), #TODO: Kernel Initialization
            nn.BatchNorm1d(out_channels),
            nn.ReLU(),
            #nn.Tanh(),
            nn.Dropout(0.2)
        )

        self.conv2 = nn.Conv1d(out_channels, out_channels, kernel_size, stride=1, padding=(kernel_size//2))
        #self.conv2 = SelfONN1dlayer(out_channels, out_channels, kernel_size, stride=1, padding=(kernel_size//2), q=3)

        self._bn_relu = nn.Sequential(
            nn.BatchNorm1d(out_channels),
            nn.ReLU(),
            nn.Dropout(0.2)
        )


    def forward(self, input):
        shortcut = self.shortcut(input)
        input = self.conv1(input)
        input = self.conv2(input)
        input = input + shortcut
        input = self._bn_relu(input)
        return input


class ResNet18(nn.Module):
    def __init__(self, in_channels, resblock):
        super().__init__()

        self.layer0 = nn.Sequential(
            nn.Conv1d(in_channels, 16, kernel_size=7, stride=1, padding=3), #TODO: Kernel Initialization
            nn.BatchNorm1d(16),
            nn.ReLU()
        )
        self.layer1 = resblock(in_channels = 16, out_channels = 16, kernel_size = 7, subsample_length = 1)
        self.layer2 = resblock(in_channels = 16, out_channels = 16, kernel_size = 7, subsample_length = 4)
        self.layer3 = resblock(in_channels = 16, out_channels = 16, kernel_size = 7, subsample_length = 1)
        self.layer4 = resblock(in_channels = 16, out_channels = 16, kernel_size = 7, subsample_length = 4)
        self.layer5 = resblock(in_channels = 16, out_channels = 16, kernel_size = 5, subsample_length = 1)
        self.layer6 = resblock(in_channels = 16, out_channels = 16, kernel_size = 5, subsample_length = 4)
        self.layer7 = resblock(in_channels = 16, out_channels = 16, kernel_size = 3, subsample_length = 1)
        self.layer8 = resblock(in_channels = 16, out_channels = 16, kernel_size = 3, subsample_length = 4)

        self.gap = torch.nn.AdaptiveAvgPool1d(1)

        self.fc_murmur =  torch.nn.Linear(16, 3)
        self.fc_outcome = torch.nn.Linear(16, 2)

    def forward(self, input):
        input = self.layer0(input)
        input = self.layer1(input)
        input = self.layer2(input)
        input = self.layer3(input)
        input = self.layer4(input)
        input = self.layer5(input)
        input = self.layer6(input)
        input = self.layer7(input)
        input = self.layer8(input)

        input = self.gap(input)
        feats = input.view((-1, input.shape[1]))
        #input = self.fc(input)
        output_murmur = self.fc_murmur(feats)
        output_outcome = self.fc_outcome(feats)

        return output_murmur, output_outcome, feats


if __name__ == "__main__":
    X = torch.randn(1,1,4096)
    resnet18 = ResNet18(1, ResBlock)
    y = resnet18(X)
